! WHIZARD 2.2.0_alpha-4 Nov 15 2013
! 
! Copyright (C) 1999-2013 by 
!     Wolfgang Kilian <kilian@physik.uni-siegen.de>
!     Thorsten Ohl <ohl@physik.uni-wuerzburg.de>
!     Juergen Reuter <juergen.reuter@desy.de>
!     
!     with contributions by 
!     Christian Speckner <cnspeckn@googlemail.com> 
!     and  Felix Braam, Sebastian Schmidt, Daniel Wiesler 
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module sf_epa

  use kinds, only: default !NODEP!
  use iso_varying_string, string_t => varying_string !NODEP!
  use constants, only: pi !NODEP!
  use file_utils !NODEP!
  use diagnostics !NODEP!
  use lorentz !NODEP!
  use unit_tests
  use os_interface
  use models
  use flavors
  use colors
  use quantum_numbers
  use state_matrices
  use polarizations
  use interactions
  use sf_aux
  use sf_base
  
  implicit none
  private

  public :: epa_data_t
  public :: sf_epa_test

  integer, parameter :: NONE = 0
  integer, parameter :: ZERO_QMIN = 1
  integer, parameter :: Q_MAX_TOO_SMALL = 2
  integer, parameter :: ZERO_XMIN = 3

  type, extends(sf_data_t) :: epa_data_t
     private
     type(model_t), pointer :: model => null ()
     type(flavor_t) :: flv
     real(default) :: alpha
     real(default) :: x_min
     real(default) :: x_max
     real(default) :: q_min
     real(default) :: q_max
     real(default) :: E_max
     real(default) :: mass
     real(default) :: log
     real(default) :: c0
     real(default) :: c1
     integer :: error = NONE
     logical :: recoil = .false.
     logical :: map = .true.
   contains
     procedure :: init => epa_data_init
     procedure :: check => epa_data_check
     procedure :: write => epa_data_write
     procedure :: get_n_par => epa_data_get_n_par
     procedure :: get_pdg_out => epa_data_get_pdg_out
     procedure :: allocate_sf_int => epa_data_allocate_sf_int  
  end type epa_data_t

  !!! JRR: WK please check
  !!! q Schmeiss raus
  type, extends (sf_int_t) :: epa_t
     type(epa_data_t), pointer :: data => null ()
     real(default) :: x = 0
     real(default) :: q = 0     
   contains
     procedure :: type_string => epa_type_string
     procedure :: write => epa_write
     procedure :: complete_kinematics => epa_complete_kinematics
     procedure :: inverse_kinematics => epa_inverse_kinematics
     procedure :: init => epa_init
     procedure :: apply => epa_apply
  end type epa_t 
  

contains

  subroutine epa_data_init &
       (data, model, flv, alpha, x_min, q_min, E_max, mass, recoil)
    class(epa_data_t), intent(inout) :: data
    type(model_t), intent(in), target :: model
    type(flavor_t), intent(in) :: flv
    real(default), intent(in) :: alpha, x_min, q_min, E_max
    real(default), intent(in), optional :: mass
    logical, intent(in), optional :: recoil
    data%model => model
    data%flv = flv
    data%alpha = alpha
    data%E_max = E_max
    data%x_min = x_min
    data%x_max = 1
    if (data%x_min == 0) then
       data%error = ZERO_XMIN;  return
    end if
    data%q_min = q_min
    data%q_max = 2 * data%E_max
    select case (char (model_get_name (data%model)))
    case ("QCD","Test")
       call msg_fatal ("EPA structure function not available for model " &
            // char (model_get_name (data%model)) // ".")
    end select     
    if (present (recoil)) then
       data%recoil = recoil
    end if
    if (present (mass)) then
       data%mass = mass
    else
       data%mass = flavor_get_mass (flv)
    end if
    if (max (data%mass, data%q_min) == 0) then
       data%error = ZERO_QMIN;  return
    else if (max (data%mass, data%q_min) >= data%E_max) then
       data%error = Q_MAX_TOO_SMALL;  return
    end if
    data%log = log (4 * (data%E_max / max (data%mass, data%q_min)) ** 2 )
    data%c0 = log (data%x_min) * (data%log - log (data%x_min))
    data%c1 = log (data%x_max) * (data%log - log (data%x_max))
  end subroutine epa_data_init

  subroutine epa_data_check (data)
    class(epa_data_t), intent(in) :: data
    select case (data%error)
    case (ZERO_QMIN)
       call msg_fatal (" EPA: Particle mass is zero")
    case (Q_MAX_TOO_SMALL)
       call msg_fatal (" EPA: Particle mass exceeds Qmax")
    case (ZERO_XMIN)
       call msg_fatal (" EPA: x_min must be larger than zero")
    end select
  end subroutine epa_data_check

  subroutine epa_data_write (data, unit, verbose) !, md5)
    class(epa_data_t), intent(in) :: data
    integer, intent(in), optional :: unit
    logical, intent(in), optional :: verbose
    integer :: u
    !!! JRR: please check: do we still need this?
!    logical, intent(in), optional :: md5
    u = output_unit (unit);  if (u < 0)  return
    write (u, "(1x,A)") "EPA data:"
    if (flavor_is_defined (data%flv)) then
       write (u, "(3x,A)", advance="no") "  flavor =  "
       call flavor_write (data%flv, u); write (u, *)
       write (u, "(3x,A,ES19.12)") "  alpha  = ", data%alpha
       write (u, "(3x,A,ES19.12)") "  x_min  = ", data%x_min
       write (u, "(3x,A,ES19.12)") "  x_max  = ", data%x_max
       write (u, "(3x,A,ES19.12)") "  q_min  = ", data%q_min
       write (u, "(3x,A,ES19.12)") "  q_max  = ", data%q_max
       write (u, "(3x,A,ES19.12)") "  E_max  = ", data%e_max
       write (u, "(3x,A,ES19.12)") "  mass   = ", data%mass
       write (u, "(3x,A,ES19.12)") "  log    = ", data%log
       write (u, "(3x,A,ES19.12)") "  c0     = ", data%c0
       write (u, "(3x,A,ES19.12)") "  c1     = ", data%c1
       write (u, "(3x,A,L2)")      "  recoil = ", data%recoil
       write (u, "(3x,A,L2)")      "  map    = ", data%map
    else
       write (u, "(3x,A)") "[undefined]"
    end if
  end subroutine epa_data_write

  function epa_data_get_n_par (data) result (n)
    class(epa_data_t), intent(in) :: data
    integer :: n
    if (data%recoil) then
       n = 3
    else
       n = 1
    end if
  end function epa_data_get_n_par
  
  function epa_data_get_pdg_out (data) result (pdg_out)
    class(epa_data_t), intent(in) :: data
    integer, dimension(:), allocatable :: pdg_out
    integer :: i, n
    n = 1
    allocate (pdg_out (n))
    pdg_out(1) = PHOTON
  end function epa_data_get_pdg_out
  
  subroutine epa_data_allocate_sf_int (data, sf_int)
    class(epa_data_t), intent(in) :: data
    class(sf_int_t), intent(inout), allocatable :: sf_int
    allocate (epa_t :: sf_int)
  end subroutine epa_data_allocate_sf_int
  
  function epa_type_string (object) result (string)
    class(epa_t), intent(in) :: object
    type(string_t) :: string
    if (associated (object%data)) then
       string = "EPA: equivalent photon approx." 
    else
       string = "EPA: [undefined]"
    end if
  end function epa_type_string
  
  subroutine epa_write (object, unit)
    !!! JRR: WK please check
    class(epa_t), intent(in) :: object
    integer, intent(in), optional :: unit
    integer :: u
    u = output_unit (unit)
    if (associated (object%data)) then
       call object%data%write (u)
       if (object%status >= SF_DONE_KINEMATICS) then
          write (u, "(1x,A)")  "SF parameters:"
          write (u, "(3x,A,ES17.10)")  "x =", object%x
          if (object%status >= SF_FAILED_EVALUATION) then
             write (u, "(3x,A,ES17.10)")  "Q =", object%q
          end if          
       end if
       call object%base_write (u)
    else
       write (u, "(1x,A)")  "EPA data: [undefined]"
    end if
  end subroutine epa_write
    
  subroutine epa_complete_kinematics (sf_int, x, f, r, map)
    class(epa_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(out) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(in) :: r
    logical, intent(in) :: map
    real(default) :: rb1, xb1
    real(default) :: d, lx
    if (map) then                                   
       rb1 = 1 - r(1)
       d = sf_int%data%log ** 2 -  4 * (r(1) * sf_int%data%c1 &
            + rb1 * sf_int%data%c0)
       if (d <= 0) then
          sf_int%status = SF_FAILED_KINEMATICS          
          f = 0
          return
       else
          lx = (sf_int%data%log - sqrt (d)) / 2
       end if
       x(1) = exp (lx)
       f = x(1) * (sf_int%data%c1 - sf_int%data%c0) / (sf_int%data%log - 2 * lx)
    else
       x(1) = r(1)
       if (sf_int%data%x_min < x(1) .and. x(1) < sf_int%data%x_max) then
          f = 1
       else
          sf_int%status = SF_FAILED_KINEMATICS
          f = 0
          return
       end if       
    end if       
    xb1 = 1 - x(1)
    if (size(x) == 3)  x(2:3) = r(2:3)
    sf_int%data%map = map
    call sf_int%split_momentum (x, xb1)
    select case (sf_int%status)
    case (SF_DONE_KINEMATICS)
       sf_int%x = x(1)
    case (SF_FAILED_KINEMATICS)
       sf_int%x = 0
       f = 0
    end select
  end subroutine epa_complete_kinematics

  subroutine epa_inverse_kinematics (sf_int, x, f, r, map, set_momenta)
    class(epa_t), intent(inout) :: sf_int
    real(default), dimension(:), intent(in) :: x
    real(default), intent(out) :: f
    real(default), dimension(:), intent(out) :: r
    logical, intent(in) :: map
    logical, intent(in), optional :: set_momenta
    real(default) :: xb1
    real(default) :: lx
    logical :: set_mom
    set_mom = .false.;  if (present (set_momenta))  set_mom = set_momenta
    if (map) then
       lx = log (x(1))
       r(1) = (lx * (sf_int%data%log - lx) - sf_int%data%c0) / &
            (sf_int%data%c1 - sf_int%data%c0)
       f = x(1) * (sf_int%data%c1 - sf_int%data%c0) / (sf_int%data%log - 2 * lx)
    else
       r(1) = x(1)                          
       if (sf_int%data%x_min < x(1) .and. x(1) < sf_int%data%x_max) then
          f = 1
       else
          f = 0
       end if
    end if
    xb1 = 1 - x(1)
    if (size(r) == 3)  r(2:3) = x(2:3)
    if (set_mom) then
       call sf_int%split_momentum (x, xb1)
       select case (sf_int%status)
       case (SF_DONE_KINEMATICS)
          sf_int%x = x(1)
       case (SF_FAILED_KINEMATICS)
          sf_int%x = 0
          f = 0
       end select
    end if
  end subroutine epa_inverse_kinematics

  subroutine epa_init (sf_int, data)
    class(epa_t), intent(out) :: sf_int
    class(sf_data_t), intent(in), target :: data
    type(quantum_numbers_mask_t), dimension(3) :: mask
    integer, dimension(3) :: hel_lock
    type(polarization_t) :: pol
    type(quantum_numbers_t), dimension(1) :: qn_fc, qn_hel
    type(flavor_t) :: flv_photon
    type(quantum_numbers_t) :: qn_photon, qn
    type(state_iterator_t) :: it_hel
    mask = new_quantum_numbers_mask (.false., .false., &
         mask_h = (/ .false., .false., .true. /))
    hel_lock = (/ 2, 1, 0 /)
    select type (data)
    type is (epa_data_t)
       call sf_int%base_init (mask, [data%mass**2], &
            [data%mass**2], [0._default], hel_lock = hel_lock)       
       sf_int%data => data
       call flavor_init (flv_photon, PHOTON, data%model)
       call quantum_numbers_init (qn_photon, flv_photon)
       call polarization_init_generic (pol, data%flv)
       call quantum_numbers_init (qn_fc(1), &
            flv = data%flv, col = color_from_flavor (data%flv, 1))
       call state_iterator_init (it_hel, pol%state)
       do while (state_iterator_is_valid (it_hel))
          qn_hel = state_iterator_get_quantum_numbers (it_hel)
          qn = qn_hel(1) .merge. qn_fc(1)
          call interaction_add_state (sf_int%interaction_t, &
               (/ qn, qn, qn_photon /))
          call state_iterator_advance (it_hel)
       end do
       call polarization_final (pol)
       call interaction_freeze (sf_int%interaction_t)
       call sf_int%set_incoming ([1])
       call sf_int%set_radiated ([2])
       call sf_int%set_outgoing ([3])
       sf_int%status = SF_INITIAL
    end select
  end subroutine epa_init
    
  subroutine epa_apply (sf_int, scale) !!!r, epa_data, no_map)
    !!! JRR: WK please check
    class(epa_t), intent(inout) :: sf_int
    real(default), intent(in) :: scale
    real(default) :: r, rb
    real(default) :: x, xb, qminsq, qmaxsq, f
    associate (data => sf_int%data)
      x = sf_int%x
      f = 0
      xb = 1 - x
      !!! JRR: WK please check: should be correct. The structure
      !!! function depends on mapped variables in the mapped case
      !!! and the true r values (constrained to interval [xmin, xmax])
      !!! for the unmapped case. Jacobian is taken into account in
      !!! complete_kinematics. Is this true also for the qminsq bound
      !!! or do we have to use the unmapped x values there?
      qminsq = max (x ** 2 / xb * data%mass ** 2, data%q_min ** 2)
      qmaxsq = min (4 * scale ** 2, data%q_max ** 2)
      if (qminsq < qmaxsq) then
         f = (data%alpha / PI) * flavor_get_charge (data%flv)**2 * &
              ((xb + x ** 2 / 2) * log (qmaxsq / qminsq) &
              - (1 - x / 2) ** 2 * log &
              ((x**2 + qmaxsq / scale ** 2) / (x**2 + qminsq / scale ** 2)) &
              - x ** 2 * data%mass ** 2 / qminsq * (1 - qminsq / qmaxsq))
      end if
    end associate
    call interaction_set_matrix_element &
         (sf_int%interaction_t, cmplx (f, kind=default))
    sf_int%status = SF_EVALUATED
  end subroutine epa_apply


  subroutine sf_epa_test (u, results)
    integer, intent(in) :: u
    type(test_results_t), intent(inout) :: results
    !!! JRR: WK please check
    !!! there are probably more tests necessary. e.g. to see whether
    !!! integrals with and without mapping are the same, right?
    call test (sf_epa_1, "sf_epa_1", &
         "structure function configuration", &
         u, results)
    call test (sf_epa_2, "sf_epa_2", &
         "structure function instance", &
         u, results)
  end subroutine sf_epa_test
  
  subroutine sf_epa_1 (u)
    integer, intent(in) :: u
    type(os_data_t) :: os_data
    type(model_list_t) :: model_list
    type(model_t), pointer :: model
    type(flavor_t) :: flv
    class(sf_data_t), allocatable :: data
    
    write (u, "(A)")  "* Test output: sf_epa_1"
    write (u, "(A)")  "*   Purpose: initialize and display &
         &test structure function data"
    write (u, "(A)")
    
    write (u, "(A)")  "* Create empty data object"
    write (u, "(A)")

    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list%read_model (var_str ("QED"), &
         var_str ("QED.mdl"), os_data, model)
    call flavor_init (flv, ELECTRON, model)

    allocate (epa_data_t :: data)
    call data%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Initialize"
    write (u, "(A)")

    select type (data)
    type is (epa_data_t)
       call data%init (model, flv, 1./137._default, 0.01_default, &
            10._default, 50._default, 0.000511_default, recoil = .false.)
    end select

    call data%write (u)

    write (u, "(A)")

    write (u, "(1x,A)")  "Outgoing particle codes:"
    write (u, "(2x,99(1x,I0))")  data%get_pdg_out ()
        
    call model_list%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: sf_epa_1"

  end subroutine sf_epa_1

  subroutine sf_epa_2 (u)
    integer, intent(in) :: u
    type(os_data_t) :: os_data
    type(model_list_t) :: model_list
    type(model_t), pointer :: model
    type(flavor_t) :: flv
    class(sf_data_t), allocatable, target :: data
    class(sf_int_t), allocatable :: sf_int
    type(vector4_t) :: k
    type(vector4_t), dimension(2) :: q
    real(default) :: E
    real(default), dimension(:), allocatable :: r, x
    real(default) :: f, s
    
    write (u, "(A)")  "* Test output: sf_epa_2"
    write (u, "(A)")  "*   Purpose: initialize and fill &
         &test structure function object"
    write (u, "(A)")
    
    write (u, "(A)")  "* Initialize configuration data"
    write (u, "(A)")

    call os_data_init (os_data)
    call syntax_model_file_init ()
    call model_list%read_model (var_str ("QED"), &
         var_str ("QED.mdl"), os_data, model)
    call flavor_init (flv, ELECTRON, model)

    call reset_interaction_counter ()
    
    allocate (epa_data_t :: data)
    select type (data)
    type is (epa_data_t)
       call data%init (model, flv, 1./137._default, 0.01_default, &
            10._default, 50._default, 0.000511_default, recoil = .false.)
    end select
       
    write (u, "(A)")  "* Initialize structure-function object"
    write (u, "(A)")
    
    call data%allocate_sf_int (sf_int)
    call sf_int%init (data)
    call sf_int%set_beam_index ([1])
    
    call sf_int%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Initialize incoming momentum with E=500"
    write (u, "(A)")
    E = 500
    k = vector4_moving (E, sqrt (E**2 - flavor_get_mass (flv)**2), 3)
    call pacify (k, 1e-10_default)
    call vector4_write (k, u)
    call sf_int%seed_kinematics ([k])
    
    write (u, "(A)")
    write (u, "(A)")  "* Set kinematics for x=0.5, no EPA mapping, collinear"
    write (u, "(A)")
    
    allocate (r (data%get_n_par ()))
    allocate (x (size (r)))
    
    r = 0.5_default
    call sf_int%complete_kinematics (x, f, r, map=.false.)
    call sf_int%write (u)
    
    write (u, "(A)")
    write (u, "(A,9(1x,F12.9))")  "x =", x
    write (u, "(A,9(1x,F12.9))")  "f =", f
    
    write (u, "(A)")
    write (u, "(A)")  "* Recover x from momenta"
    write (u, "(A)")
    
    q = interaction_get_momenta (sf_int%interaction_t, outgoing=.true.)
    call sf_int%final ()
    deallocate (sf_int)
    
    call data%allocate_sf_int (sf_int)
    call sf_int%init (data)
    call sf_int%set_beam_index ([1])
    
    call sf_int%seed_kinematics ([k])
    call interaction_set_momenta (sf_int%interaction_t, q, outgoing=.true.)
    call sf_int%recover_x (x)
    
    write (u, "(A,9(1x,F12.9))")  "x =", x
    
    write (u, "(A)")
    write (u, "(A)")  "* Evaluate EPA structure function"
    write (u, "(A)")
    
    call sf_int%complete_kinematics (x, f, r, map=.false.) 
    call sf_int%apply (scale = 100._default)
    call sf_int%write (u)
    
    call sf_int%final ()
    deallocate (sf_int)
    deallocate (x)
    deallocate (r)
    deallocate (data)

    write (u, "(A)")
    write (u, "(A)")  "* Initialize structure-function object"
    write (u, "(A)")

    allocate (epa_data_t :: data)
    select type (data)
    type is (epa_data_t)
       call data%init (model, flv, 1./137._default, 0.01_default, &
            10._default, 50._default, 0.000511_default, recoil = .true.)
    end select    
    
    call data%allocate_sf_int (sf_int)
    call sf_int%init (data)
    call sf_int%set_beam_index ([1])
    
    call sf_int%write (u)

    write (u, "(A)")
    write (u, "(A)")  "* Initialize incoming momentum with E=500"
    write (u, "(A)")
    E = 500
    k = vector4_moving (E, sqrt (E**2 - flavor_get_mass (flv)**2), 3)
    call pacify (k, 1e-10_default)
    call vector4_write (k, u)
    call sf_int%seed_kinematics ([k])
        
    write (u, "(A)")
    write (u, "(A)")  "* Set kinematics for x=0.5/0.5/0.25, with EPA mapping, "
    write (u, "(A)")  "          non-coll., keeping energy"
    write (u, "(A)")
    
    allocate (r (data%get_n_par ()))
    allocate (x (size (r)))
    
    r = [0.5_default, 0.5_default, 0.25_default]
    sf_int%on_shell_mode = KEEP_ENERGY    
    call sf_int%complete_kinematics (x, f, r, map=.true.)
    call interaction_pacify_momenta (sf_int%interaction_t, 1e-10_default)
    call sf_int%write (u)
    
    write (u, "(A)")
    write (u, "(A,9(1x,F12.9))")  "x =", x
    write (u, "(A,9(1x,F12.9))")  "f =", f
    
    write (u, "(A)")
    write (u, "(A)")  "* Recover x and r from momenta"
    write (u, "(A)")
    
    q = interaction_get_momenta (sf_int%interaction_t, outgoing=.true.)
    call sf_int%final ()
    deallocate (sf_int)
    
    call data%allocate_sf_int (sf_int)
    call sf_int%init (data)
    call sf_int%set_beam_index ([1])
    
    call sf_int%seed_kinematics ([k])
    call interaction_set_momenta (sf_int%interaction_t, q, outgoing=.true.)
    call sf_int%recover_x (x)
    call sf_int%inverse_kinematics (x, f, r, map=.true.)    
    
    write (u, "(A,9(1x,F12.9))")  "x =", x
    write (u, "(A,9(1x,F12.9))")  "r =", r
    
    write (u, "(A)")
    write (u, "(A)")  "* Evaluate EPA structure function"
    write (u, "(A)")
    
    call sf_int%complete_kinematics (x, f, r, map=.true.) 
    call interaction_pacify_momenta (sf_int%interaction_t, 1e-10_default)    
    call sf_int%apply (scale = 100._default)
    call sf_int%write (u)
    
    
    write (u, "(A)")
    write (u, "(A)")  "* Cleanup"
    
    call sf_int%final ()
    call model_list%final ()
    call syntax_model_file_final ()
    
    write (u, "(A)")
    write (u, "(A)")  "* Test output end: sf_epa_2"

  end subroutine sf_epa_2


end module sf_epa
